from .VixHandle import VixHandle
from .VixError import VixError
from vix import _backend
vix = _backend._vix
ffi = _backend._ffi


class VixSnapshot(VixHandle):
    """Represents a VM's snapshot"""

    def __init__(self, handle):
        super(VixSnapshot, self).__init__(handle)
        assert self.get_type() == VixHandle.VIX_HANDLETYPE_SNAPSHOT, 'Expected VixSnapshot handle.'

    @property
    def name(self):
        """Gets the snapshot's name.

        :rtype: str
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_SNAPSHOT_DISPLAYNAME)

    @property
    def description(self):
        """Get the snapshot's description.

        :rtype: str
        """
        return self.get_properties(VixHandle.VIX_PROPERTY_SNAPSHOT_DESCRIPTION)
    
    @property
    def power_state(self):
        """Gets the snapshot's power state.
        
        :returns: Power state, can be a combination of VixVM.VIX_VM_POWERSTATE_*
        :rtype: int
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_SNAPSHOT_POWERSTATE)

    def get_num_children(self):
        """Gets the number of children the current snapshot has.

        :raises vix.VixError: On failure to get child count.
        """

        child_count = ffi.new('int*')
        error_code = vix.VixSnapshot_GetNumChildren(
            self._handle,
            child_count,
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        return int(child_count[0])

    def get_child(self, child_index):
        """Gets a child snapshot at the designated index

        :param int child_index: Index of child snapshot.

        :returns: Snapshot at specified index.
        :rtype: .VixSnapshot

        :raises vix.VixError: On failure to retrieve snapshot.
        """

        child_handle = ffi.new('VixHandle*')
        error_code = vix.VixSnapshot_GetChild(
            self._handle,
            child_index,
            child_handle,
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        return VixSnapshot(child_handle[0])

    def get_parent(self):
        """Gets the parent of the current snapshot.

        :returns: Parent of current snapshot or None if snapshot is a root.
        :rtype: .VixSnapshot

        :raises vix.VixError: On failure to get snapshot.
        """
        
        parent_handle = ffi.new('VixHandle*')
        error_code = vix.VixSnapshot_GetParent(
            self._handle,
            parent_handle,
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        temp_handle = VixHandle(parent_handle[0])
        if temp_handle.get_type() == VixHandle.VIX_HANDLETYPE_NONE:
            temp_handle.release()
            return None

        return VixSnapshot(parent_handle[0])



from __future__ import absolute_import

from .compat import _str
from vix import _backend, API_ENCODING
vix = _backend._vix
ffi = _backend._ffi

class VixError(Exception):
    VIX_OK = 0

    # General errors
    VIX_E_FAIL = 1
    VIX_E_OUT_OF_MEMORY = 2
    VIX_E_INVALID_ARG = 3
    VIX_E_FILE_NOT_FOUND = 4
    VIX_E_OBJECT_IS_BUSY = 5
    VIX_E_NOT_SUPPORTED = 6
    VIX_E_FILE_ERROR = 7
    VIX_E_DISK_FULL = 8
    VIX_E_INCORRECT_FILE_TYPE = 9
    VIX_E_CANCELLED = 10
    VIX_E_FILE_READ_ONLY = 11
    VIX_E_FILE_ALREADY_EXISTS = 12
    VIX_E_FILE_ACCESS_ERROR = 13
    VIX_E_REQUIRES_LARGE_FILES = 14
    VIX_E_FILE_ALREADY_LOCKED = 15
    VIX_E_VMDB = 16
    VIX_E_NOT_SUPPORTED_ON_REMOTE_OBJECT = 20
    VIX_E_FILE_TOO_BIG = 21
    VIX_E_FILE_NAME_INVALID = 22
    VIX_E_ALREADY_EXISTS = 23
    VIX_E_BUFFER_TOOSMALL = 24
    VIX_E_OBJECT_NOT_FOUND = 25
    VIX_E_HOST_NOT_CONNECTED = 26
    VIX_E_INVALID_UTF8_STRING = 27
    VIX_E_OPERATION_ALREADY_IN_PROGRESS = 31
    VIX_E_UNFINISHED_JOB = 29
    VIX_E_NEED_KEY = 30
    VIX_E_LICENSE = 32
    VIX_E_VM_HOST_DISCONNECTED = 34
    VIX_E_AUTHENTICATION_FAIL = 35
    VIX_E_HOST_CONNECTION_LOST = 36
    VIX_E_DUPLICATE_NAME = 41
    VIX_E_ARGUMENT_TOO_BIG = 44

    # Handle Errors
    VIX_E_INVALID_HANDLE = 1000
    VIX_E_NOT_SUPPORTED_ON_HANDLE_TYPE = 1001
    VIX_E_TOO_MANY_HANDLES = 1002

    # XML errors
    VIX_E_NOT_FOUND = 2000
    VIX_E_TYPE_MISMATCH = 2001
    VIX_E_INVALID_XML = 2002

    # VM Control Errors
    VIX_E_TIMEOUT_WAITING_FOR_TOOLS = 3000
    VIX_E_UNRECOGNIZED_COMMAND = 3001
    VIX_E_OP_NOT_SUPPORTED_ON_GUEST = 3003
    VIX_E_PROGRAM_NOT_STARTED = 3004
    VIX_E_CANNOT_START_READ_ONLY_VM = 3005
    VIX_E_VM_NOT_RUNNING = 3006
    VIX_E_VM_IS_RUNNING = 3007
    VIX_E_CANNOT_CONNECT_TO_VM = 3008
    VIX_E_POWEROP_SCRIPTS_NOT_AVAILABLE = 3009
    VIX_E_NO_GUEST_OS_INSTALLED = 3010
    VIX_E_VM_INSUFFICIENT_HOST_MEMORY = 3011
    VIX_E_SUSPEND_ERROR = 3012
    VIX_E_VM_NOT_ENOUGH_CPUS = 3013
    VIX_E_HOST_USER_PERMISSIONS = 3014
    VIX_E_GUEST_USER_PERMISSIONS = 3015
    VIX_E_TOOLS_NOT_RUNNING = 3016
    VIX_E_GUEST_OPERATIONS_PROHIBITED = 3017
    VIX_E_ANON_GUEST_OPERATIONS_PROHIBITED = 3018
    VIX_E_ROOT_GUEST_OPERATIONS_PROHIBITED = 3019
    VIX_E_MISSING_ANON_GUEST_ACCOUNT = 3023
    VIX_E_CANNOT_AUTHENTICATE_WITH_GUEST = 3024
    VIX_E_UNRECOGNIZED_COMMAND_IN_GUEST = 3025
    VIX_E_CONSOLE_GUEST_OPERATIONS_PROHIBITED = 3026
    VIX_E_MUST_BE_CONSOLE_USER = 3027
    VIX_E_VMX_MSG_DIALOG_AND_NO_UI = 3028
    VIX_E_OPERATION_NOT_ALLOWED_FOR_LOGIN_TYPE = 3031
    VIX_E_LOGIN_TYPE_NOT_SUPPORTED = 3032
    VIX_E_EMPTY_PASSWORD_NOT_ALLOWED_IN_GUEST = 3033
    VIX_E_INTERACTIVE_SESSION_NOT_PRESENT = 3034
    VIX_E_INTERACTIVE_SESSION_USER_MISMATCH = 3035
    VIX_E_CANNOT_POWER_ON_VM = 3041
    VIX_E_NO_DISPLAY_SERVER = 3043
    VIX_E_TOO_MANY_LOGONS = 3046
    VIX_E_INVALID_AUTHENTICATION_SESSION = 3047

    # VM Errors
    VIX_E_VM_NOT_FOUND = 4000
    VIX_E_NOT_SUPPORTED_FOR_VM_VERSION = 4001
    VIX_E_CANNOT_READ_VM_CONFIG = 4002
    VIX_E_TEMPLATE_VM = 4003
    VIX_E_VM_ALREADY_LOADED = 4004
    VIX_E_VM_ALREADY_UP_TO_DATE = 4006
    VIX_E_VM_UNSUPPORTED_GUEST = 4011

    # Property Errors
    VIX_E_UNRECOGNIZED_PROPERTY = 6000
    VIX_E_INVALID_PROPERTY_VALUE = 6001
    VIX_E_READ_ONLY_PROPERTY = 6002
    VIX_E_MISSING_REQUIRED_PROPERTY = 6003
    VIX_E_INVALID_SERIALIZED_DATA = 6004
    VIX_E_PROPERTY_TYPE_MISMATCH = 6005

    # Completion Errors
    VIX_E_BAD_VM_INDEX = 8000

    # Message errors
    VIX_E_INVALID_MESSAGE_HEADER = 10000
    VIX_E_INVALID_MESSAGE_BODY = 10001

    # Snapshot errors
    VIX_E_SNAPSHOT_INVAL = 13000
    VIX_E_SNAPSHOT_DUMPER = 13001
    VIX_E_SNAPSHOT_DISKLIB = 13002
    VIX_E_SNAPSHOT_NOTFOUND = 13003
    VIX_E_SNAPSHOT_EXISTS = 13004
    VIX_E_SNAPSHOT_VERSION = 13005
    VIX_E_SNAPSHOT_NOPERM = 13006
    VIX_E_SNAPSHOT_CONFIG = 13007
    VIX_E_SNAPSHOT_NOCHANGE = 13008
    VIX_E_SNAPSHOT_CHECKPOINT = 13009
    VIX_E_SNAPSHOT_LOCKED = 13010
    VIX_E_SNAPSHOT_INCONSISTENT = 13011
    VIX_E_SNAPSHOT_NAMETOOLONG = 13012
    VIX_E_SNAPSHOT_VIXFILE = 13013
    VIX_E_SNAPSHOT_DISKLOCKED = 13014
    VIX_E_SNAPSHOT_DUPLICATEDDISK = 13015
    VIX_E_SNAPSHOT_INDEPENDENTDISK = 13016
    VIX_E_SNAPSHOT_NONUNIQUE_NAME = 13017
    VIX_E_SNAPSHOT_MEMORY_ON_INDEPENDENT_DISK = 13018
    VIX_E_SNAPSHOT_MAXSNAPSHOTS = 13019
    VIX_E_SNAPSHOT_MIN_FREE_SPACE = 13020
    VIX_E_SNAPSHOT_HIERARCHY_TOODEEP = 13021
    VIX_E_SNAPSHOT_NOT_REVERTABLE = 13024

    # Host Errors
    VIX_E_HOST_DISK_INVALID_VALUE = 14003
    VIX_E_HOST_DISK_SECTORSIZE = 14004
    VIX_E_HOST_FILE_ERROR_EOF = 14005
    VIX_E_HOST_NETBLKDEV_HANDSHAKE = 14006
    VIX_E_HOST_SOCKET_CREATION_ERROR = 14007
    VIX_E_HOST_SERVER_NOT_FOUND = 14008
    VIX_E_HOST_NETWORK_CONN_REFUSED = 14009
    VIX_E_HOST_TCP_SOCKET_ERROR = 14010
    VIX_E_HOST_TCP_CONN_LOST = 14011
    VIX_E_HOST_NBD_HASHFILE_VOLUME = 14012
    VIX_E_HOST_NBD_HASHFILE_INIT = 14013

    # Disklib errors
    VIX_E_DISK_INVAL = 16000
    VIX_E_DISK_NOINIT = 16001
    VIX_E_DISK_NOIO = 16002
    VIX_E_DISK_PARTIALCHAIN = 16003
    VIX_E_DISK_NEEDSREPAIR = 16006
    VIX_E_DISK_OUTOFRANGE = 16007
    VIX_E_DISK_CID_MISMATCH = 16008
    VIX_E_DISK_CANTSHRINK = 16009
    VIX_E_DISK_PARTMISMATCH = 16010
    VIX_E_DISK_UNSUPPORTEDDISKVERSION = 16011
    VIX_E_DISK_OPENPARENT = 16012
    VIX_E_DISK_NOTSUPPORTED = 16013
    VIX_E_DISK_NEEDKEY = 16014
    VIX_E_DISK_NOKEYOVERRIDE = 16015
    VIX_E_DISK_NOTENCRYPTED = 16016
    VIX_E_DISK_NOKEY = 16017
    VIX_E_DISK_INVALIDPARTITIONTABLE = 16018
    VIX_E_DISK_NOTNORMAL = 16019
    VIX_E_DISK_NOTENCDESC = 16020
    VIX_E_DISK_NEEDVMFS = 16022
    VIX_E_DISK_RAWTOOBIG = 16024
    VIX_E_DISK_TOOMANYOPENFILES = 16027
    VIX_E_DISK_TOOMANYREDO = 16028
    VIX_E_DISK_RAWTOOSMALL = 16029
    VIX_E_DISK_INVALIDCHAIN = 16030
    VIX_E_DISK_KEY_NOTFOUND = 16052
    VIX_E_DISK_SUBSYSTEM_INIT_FAIL = 16053
    VIX_E_DISK_INVALID_CONNECTION = 16054
    VIX_E_DISK_ENCODING = 16061
    VIX_E_DISK_CANTREPAIR = 16062
    VIX_E_DISK_INVALIDDISK = 16063
    VIX_E_DISK_NOLICENSE = 16064
    VIX_E_DISK_NODEVICE = 16065
    VIX_E_DISK_UNSUPPORTEDDEVICE = 16066
    VIX_E_DISK_CAPACITY_MISMATCH = 16067
    VIX_E_DISK_PARENT_NOTALLOWED = 16068
    VIX_E_DISK_ATTACH_ROOTLINK = 16069

    # Crypto Library Errors
    VIX_E_CRYPTO_UNKNOWN_ALGORITHM = 17000
    VIX_E_CRYPTO_BAD_BUFFER_SIZE = 17001
    VIX_E_CRYPTO_INVALID_OPERATION = 17002
    VIX_E_CRYPTO_RANDOM_DEVICE = 17003
    VIX_E_CRYPTO_NEED_PASSWORD = 17004
    VIX_E_CRYPTO_BAD_PASSWORD = 17005
    VIX_E_CRYPTO_NOT_IN_DICTIONARY = 17006
    VIX_E_CRYPTO_NO_CRYPTO = 17007
    VIX_E_CRYPTO_ERROR = 17008
    VIX_E_CRYPTO_BAD_FORMAT = 17009
    VIX_E_CRYPTO_LOCKED = 17010
    VIX_E_CRYPTO_EMPTY = 17011
    VIX_E_CRYPTO_KEYSAFE_LOCATOR = 17012

    # Remoting Errors.
    VIX_E_CANNOT_CONNECT_TO_HOST = 18000
    VIX_E_NOT_FOR_REMOTE_HOST = 18001
    VIX_E_INVALID_HOSTNAME_SPECIFICATION = 18002

    # Screen Capture Errors.
    VIX_E_SCREEN_CAPTURE_ERROR = 19000
    VIX_E_SCREEN_CAPTURE_BAD_FORMAT = 19001
    VIX_E_SCREEN_CAPTURE_COMPRESSION_FAIL = 19002
    VIX_E_SCREEN_CAPTURE_LARGE_DATA = 19003

    # Guest Errors
    VIX_E_GUEST_VOLUMES_NOT_FROZEN = 20000
    VIX_E_NOT_A_FILE = 20001
    VIX_E_NOT_A_DIRECTORY = 20002
    VIX_E_NO_SUCH_PROCESS = 20003
    VIX_E_FILE_NAME_TOO_LONG = 20004
    VIX_E_OPERATION_DISABLED = 20005

    # Tools install errors
    VIX_E_TOOLS_INSTALL_NO_IMAGE = 21000
    VIX_E_TOOLS_INSTALL_IMAGE_INACCESIBLE = 21001
    VIX_E_TOOLS_INSTALL_NO_DEVICE = 21002
    VIX_E_TOOLS_INSTALL_DEVICE_NOT_CONNECTED = 21003
    VIX_E_TOOLS_INSTALL_CANCELLED = 21004
    VIX_E_TOOLS_INSTALL_INIT_FAILED = 21005
    VIX_E_TOOLS_INSTALL_AUTO_NOT_SUPPORTED = 21006
    VIX_E_TOOLS_INSTALL_GUEST_NOT_READY = 21007
    VIX_E_TOOLS_INSTALL_ERROR = 21009
    VIX_E_TOOLS_INSTALL_ALREADY_UP_TO_DATE = 21010
    VIX_E_TOOLS_INSTALL_IN_PROGRESS = 21011
    VIX_E_TOOLS_INSTALL_IMAGE_COPY_FAILED = 21012

    # Wrapper Errors
    VIX_E_WRAPPER_WORKSTATION_NOT_INSTALLED = 22001
    VIX_E_WRAPPER_VERSION_NOT_FOUND = 22002
    VIX_E_WRAPPER_SERVICEPROVIDER_NOT_FOUND = 22003
    VIX_E_WRAPPER_PLAYER_NOT_INSTALLED = 22004
    VIX_E_WRAPPER_RUNTIME_NOT_INSTALLED = 22005
    VIX_E_WRAPPER_MULTIPLE_SERVICEPROVIDERS = 22006

    # FuseMnt errors
    VIX_E_MNTAPI_MOUNTPT_NOT_FOUND = 24000
    VIX_E_MNTAPI_MOUNTPT_IN_USE = 24001
    VIX_E_MNTAPI_DISK_NOT_FOUND = 24002
    VIX_E_MNTAPI_DISK_NOT_MOUNTED = 24003
    VIX_E_MNTAPI_DISK_IS_MOUNTED = 24004
    VIX_E_MNTAPI_DISK_NOT_SAFE = 24005
    VIX_E_MNTAPI_DISK_CANT_OPEN = 24006
    VIX_E_MNTAPI_CANT_READ_PARTS = 24007
    VIX_E_MNTAPI_UMOUNT_APP_NOT_FOUND = 24008
    VIX_E_MNTAPI_UMOUNT = 24009
    VIX_E_MNTAPI_NO_MOUNTABLE_PARTITONS = 24010
    VIX_E_MNTAPI_PARTITION_RANGE = 24011
    VIX_E_MNTAPI_PERM = 24012
    VIX_E_MNTAPI_DICT = 24013
    VIX_E_MNTAPI_DICT_LOCKED = 24014
    VIX_E_MNTAPI_OPEN_HANDLES = 24015
    VIX_E_MNTAPI_CANT_MAKE_VAR_DIR = 24016
    VIX_E_MNTAPI_NO_ROOT = 24017
    VIX_E_MNTAPI_LOOP_FAILED = 24018
    VIX_E_MNTAPI_DAEMON = 24019
    VIX_E_MNTAPI_INTERNAL = 24020
    VIX_E_MNTAPI_SYSTEM = 24021
    VIX_E_MNTAPI_NO_CONNECTION_DETAILS = 24022

    # VixMntapi errors
    VIX_E_MNTAPI_INCOMPATIBLE_VERSION = 24300
    VIX_E_MNTAPI_OS_ERROR = 24301
    VIX_E_MNTAPI_DRIVE_LETTER_IN_USE = 24302
    VIX_E_MNTAPI_DRIVE_LETTER_ALREADY_ASSIGNED = 24303
    VIX_E_MNTAPI_VOLUME_NOT_MOUNTED = 24304
    VIX_E_MNTAPI_VOLUME_ALREADY_MOUNTED = 24305
    VIX_E_MNTAPI_FORMAT_FAILURE = 24306
    VIX_E_MNTAPI_NO_DRIVER = 24307
    VIX_E_MNTAPI_ALREADY_OPENED = 24308
    VIX_E_MNTAPI_ITEM_NOT_FOUND = 24309
    VIX_E_MNTAPI_UNSUPPROTED_BOOT_LOADER = 24310
    VIX_E_MNTAPI_UNSUPPROTED_OS = 24311
    VIX_E_MNTAPI_CODECONVERSION = 24312
    VIX_E_MNTAPI_REGWRITE_ERROR = 24313
    VIX_E_MNTAPI_UNSUPPORTED_FT_VOLUME = 24314
    VIX_E_MNTAPI_PARTITION_NOT_FOUND = 24315
    VIX_E_MNTAPI_PUTFILE_ERROR = 24316
    VIX_E_MNTAPI_GETFILE_ERROR = 24317
    VIX_E_MNTAPI_REG_NOT_OPENED = 24318
    VIX_E_MNTAPI_REGDELKEY_ERROR = 24319
    VIX_E_MNTAPI_CREATE_PARTITIONTABLE_ERROR = 24320
    VIX_E_MNTAPI_OPEN_FAILURE = 24321
    VIX_E_MNTAPI_VOLUME_NOT_WRITABLE = 24322

    # Success on operation that completes asynchronously
    VIX_ASYNC = 25000

    # Async errors
    VIX_E_ASYNC_MIXEDMODE_UNSUPPORTED = 26000

    # Network Errors
    VIX_E_NET_HTTP_UNSUPPORTED_PROTOCOL = 30001
    VIX_E_NET_HTTP_URL_MALFORMAT = 30003
    VIX_E_NET_HTTP_COULDNT_RESOLVE_PROXY = 30005
    VIX_E_NET_HTTP_COULDNT_RESOLVE_HOST = 30006
    VIX_E_NET_HTTP_COULDNT_CONNECT = 30007
    VIX_E_NET_HTTP_HTTP_RETURNED_ERROR = 30022
    VIX_E_NET_HTTP_OPERATION_TIMEDOUT = 30028
    VIX_E_NET_HTTP_SSL_CONNECT_ERROR = 30035
    VIX_E_NET_HTTP_TOO_MANY_REDIRECTS = 30047
    VIX_E_NET_HTTP_TRANSFER = 30200
    VIX_E_NET_HTTP_SSL_SECURITY = 30201
    VIX_E_NET_HTTP_GENERIC = 30202

    def __init__(self, error_code):
        self._error = error_code

    def __str__(self):
        return "VixError #{0}: {1}".format(str(self._error), self.get_error_text() or "?")

    def get_error_text(self):
        """Get an error's text

        :returns: A human readable error message.
        :rtype: str
        """

        return _str(
            ffi.string(
                vix.Vix_GetErrorText(
                    ffi.cast('VixError', self._error), 
                    ffi.cast('char*', 0)
                )
            ), 
        API_ENCODING
        )



# This file is for backward compatibility with Python2.
# If you are using Python2, please do upgrade to Python3 :)
import sys


if sys.version_info.major < 3:
    def _bytes(data, encoding=None):
        if encoding:
            return str(data).encode(encoding)
        else:
            return str(data)

    def _str(data, encoding):
        return str(data).decode(encoding)

else:
    _bytes = bytes
    _str = str



from __future__ import absolute_import

from vix import _backend, VixError, API_ENCODING
from .compat import _str, _bytes
vix = _backend._vix
ffi = _backend._ffi


class VixHandle(object):
    """Represents a handle of the VIX library.

    .. note:: Internal use.
    """

    VIX_INVALID_HANDLE = 0

    VIX_HANDLETYPE_NONE = 0
    VIX_HANDLETYPE_HOST = 2
    VIX_HANDLETYPE_VM = 3
    VIX_HANDLETYPE_NETWORK = 5
    VIX_HANDLETYPE_JOB = 6
    VIX_HANDLETYPE_SNAPSHOT = 7
    VIX_HANDLETYPE_PROPERTY_LIST = 9
    VIX_HANDLETYPE_METADATA_CONTAINER = 1

    VIX_PROPERTYTYPE_ANY = 0
    VIX_PROPERTYTYPE_INTEGER = 1
    VIX_PROPERTYTYPE_STRING = 2
    VIX_PROPERTYTYPE_BOOL = 3
    VIX_PROPERTYTYPE_HANDLE = 4
    VIX_PROPERTYTYPE_INT64 = 5
    VIX_PROPERTYTYPE_BLOB = 6

    VIX_PROPERTY_NONE = 0
    VIX_PROPERTY_META_DATA_CONTAINER = 2
    VIX_PROPERTY_HOST_HOSTTYPE = 50
    VIX_PROPERTY_HOST_API_VERSION = 51
    VIX_PROPERTY_HOST_SOFTWARE_VERSION = 52
    VIX_PROPERTY_VM_NUM_VCPUS = 101
    VIX_PROPERTY_VM_VMX_PATHNAME = 103
    VIX_PROPERTY_VM_VMTEAM_PATHNAME = 105
    VIX_PROPERTY_VM_MEMORY_SIZE = 106
    VIX_PROPERTY_VM_READ_ONLY = 107
    VIX_PROPERTY_VM_NAME = 108
    VIX_PROPERTY_VM_GUESTOS = 109
    VIX_PROPERTY_VM_IN_VMTEAM = 128
    VIX_PROPERTY_VM_POWER_STATE = 129
    VIX_PROPERTY_VM_TOOLS_STATE = 152
    VIX_PROPERTY_VM_IS_RUNNING = 196
    VIX_PROPERTY_VM_SUPPORTED_FEATURES = 197
    VIX_PROPERTY_VM_SSL_ERROR = 293
    VIX_PROPERTY_FOUND_ITEM_LOCATION = 4010
    VIX_PROPERTY_SNAPSHOT_DISPLAYNAME = 4200
    VIX_PROPERTY_SNAPSHOT_DESCRIPTION = 4201
    VIX_PROPERTY_SNAPSHOT_POWERSTATE = 4205
    VIX_PROPERTY_GUEST_SHAREDFOLDERS_SHARES_PATH = 4525
    VIX_PROPERTY_VM_ENCRYPTION_PASSWORD = 7001

    def __init__(self, handle):
        self._handle = handle

    def is_valid(self):
        """Checks that the handle is valid.

        :returns: True if handle is valid, otherwise False.
        :rtype: bool
        """

        return self._handle is not None and self._handle != self.VIX_INVALID_HANDLE

    def get_type(self):
        """Gets the handle's type.

        :returns: Type of handle from VIX_HANDLETYPE_*.
        :rtype: int
        """

        return vix.Vix_GetHandleType(self._handle)

    def add_ref(self):
        """Increases the handle's refcount."""

        vix.Vix_AddRefHandle(self._handle)

    def release(self):
        """Decreases refcount or Releases the handle."""

        vix.Vix_ReleaseHandle(self._handle)

    def get_properties(self, *args):
        c_args = list()
        ret_vals = list()

        for arg in args:
            prop_type = self.get_property_type(arg)

            alloc = None
            if prop_type in (self.VIX_PROPERTYTYPE_STRING, self.VIX_PROPERTYTYPE_BLOB):
                alloc = ffi.new('char**')
            elif prop_type == self.VIX_PROPERTYTYPE_INT64:
                alloc = ffi.new('int64*')
            else:
                alloc = ffi.new('int*')

            c_args.append(ffi.cast('VixPropertyID', arg))
            c_args.append(alloc)
            ret_vals.append((alloc, prop_type))

        c_args.append(ffi.cast('VixPropertyID', self.VIX_PROPERTY_NONE))

        error_code = vix.Vix_GetProperties(self._handle, *c_args)

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        results = list()
        for ret_val in ret_vals:
            val, val_type = ret_val

            if val_type in (self.VIX_PROPERTYTYPE_INT64, self.VIX_PROPERTYTYPE_INTEGER, self.VIX_PROPERTYTYPE_HANDLE):
                results.append(int(val[0]))
            elif val_type == self.VIX_PROPERTYTYPE_STRING:
                results.append(_str(ffi.string(val[0]), API_ENCODING))
            elif val_type == self.VIX_PROPERTYTYPE_BOOL:
                results.append(bool(val[0]))
            else:
                assert False, "Can't handle property type."

        return results[0] if len(results) == 1 else results


    def get_property_type(self, property_id):
        prop_type = ffi.new('VixPropertyType*')
        error_code = vix.Vix_GetPropertyType(
            self._handle,
            ffi.cast('VixPropertyID', property_id),
            prop_type
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        return int(prop_type[0])



"""
    vix
    ---

    A python binding for VMWare's VIX library.

    :copyright: (c) 2016 by Naim A.
    :license: GPLv3, see LICENSE for more details.
"""

__version__ = "1.0.7"

from .VixBackend import VixBackend as _VixBackend
_backend = _VixBackend()
vix = _backend._vix
ffi = _backend._ffi

# should be ascii for windows, utf8 for linux
API_ENCODING = 'utf-8'

from .VixError import VixError
from .VixHost import VixHost
from .VixSnapshot import VixSnapshot
from .VixVM import VixVM



from __future__ import absolute_import

from .compat import _bytes, _str
from .VixHandle import VixHandle
from .VixError import VixError
from vix import _backend, API_ENCODING
import datetime

vix = _backend._vix
ffi = _backend._ffi

class VixJob(VixHandle):
    """Represnts a VIX Job handle.

    .. note:: Internal use.
    """

    VIX_PROPERTY_JOB_RESULT_ERROR_CODE = 3000
    VIX_PROPERTY_JOB_RESULT_VM_IN_GROUP = 3001
    VIX_PROPERTY_JOB_RESULT_USER_MESSAGE = 3002
    VIX_PROPERTY_JOB_RESULT_EXIT_CODE = 3004
    VIX_PROPERTY_JOB_RESULT_COMMAND_OUTPUT = 3005
    VIX_PROPERTY_JOB_RESULT_HANDLE = 3010
    VIX_PROPERTY_JOB_RESULT_GUEST_OBJECT_EXISTS = 3011
    VIX_PROPERTY_JOB_RESULT_GUEST_PROGRAM_ELAPSED_TIME = 3017
    VIX_PROPERTY_JOB_RESULT_GUEST_PROGRAM_EXIT_CODE = 3018
    VIX_PROPERTY_JOB_RESULT_ITEM_NAME = 3035
    VIX_PROPERTY_JOB_RESULT_FOUND_ITEM_DESCRIPTION = 3036
    VIX_PROPERTY_JOB_RESULT_SHARED_FOLDER_COUNT = 3046
    VIX_PROPERTY_JOB_RESULT_SHARED_FOLDER_HOST = 3048
    VIX_PROPERTY_JOB_RESULT_SHARED_FOLDER_FLAGS = 3049
    VIX_PROPERTY_JOB_RESULT_PROCESS_ID = 3051
    VIX_PROPERTY_JOB_RESULT_PROCESS_OWNER = 3052
    VIX_PROPERTY_JOB_RESULT_PROCESS_COMMAND = 3053
    VIX_PROPERTY_JOB_RESULT_FILE_FLAGS = 3054
    VIX_PROPERTY_JOB_RESULT_PROCESS_START_TIME = 3055
    VIX_PROPERTY_JOB_RESULT_VM_VARIABLE_STRING = 3056
    VIX_PROPERTY_JOB_RESULT_PROCESS_BEING_DEBUGGED = 3057
    VIX_PROPERTY_JOB_RESULT_SCREEN_IMAGE_SIZE = 3058
    VIX_PROPERTY_JOB_RESULT_SCREEN_IMAGE_DATA = 3059
    VIX_PROPERTY_JOB_RESULT_FILE_SIZE = 3061
    VIX_PROPERTY_JOB_RESULT_FILE_MOD_TIME = 3062
    VIX_PROPERTY_JOB_RESULT_EXTRA_ERROR_INFO = 3084

    VIX_FILE_ATTRIBUTES_DIRECTORY = 0x0001
    VIX_FILE_ATTRIBUTES_SYMLINK = 0x0002

    STR_RESULT_TYPES = (
        VIX_PROPERTY_JOB_RESULT_ITEM_NAME,
        VIX_PROPERTY_JOB_RESULT_VM_VARIABLE_STRING,
        VIX_PROPERTY_JOB_RESULT_COMMAND_OUTPUT,
        VIX_PROPERTY_JOB_RESULT_PROCESS_OWNER,
        VIX_PROPERTY_JOB_RESULT_PROCESS_COMMAND,
        VIX_PROPERTY_JOB_RESULT_SHARED_FOLDER_HOST,
    )

    def __init__(self, handle):
        super(VixJob, self).__init__(handle)
        assert self.get_type() == VixHandle.VIX_HANDLETYPE_JOB, 'Expected VixJob handle.'

    def wait(self, *args):
        """Waits for the job to complete and gets requested results.

        :param \*args: A list of properties to retreive (VIX_PROPERTY_JOB_RESULT_*).

        :returns: A tuple of results if more than one object was requested.

        :raises vix.VixError: If job failed.
        """

        c_args = list()
        ret_data = list()

        for arg in args:
            c_args.append(ffi.cast('VixPropertyType', arg))

            # TODO: Check the arg type and allocate accordingly...
            alloc = None
            if arg in self.STR_RESULT_TYPES:
                alloc = ffi.new('char**')
            else:
                alloc = ffi.new('int*')
            ret_data.append(alloc)
            c_args.append(alloc)

        c_args.append(ffi.cast('VixPropertyType', self.VIX_PROPERTY_NONE))

        error_code = vix.VixJob_Wait(self._handle, *c_args)
        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        # deref data...
        result = list()
        for i in range(len(args)):
            if args[i] == self.VIX_PROPERTY_NONE:
                break
            val = ret_data[i]
            if args[i] in self.STR_RESULT_TYPES:
                result.append(_str(ffi.string(val[0]), API_ENCODING))
                vix.Vix_FreeBuffer(val[0])
            else:
                result.append(val[0])

        return result[0] if len(result) == 1 else result

    def is_done(self):
        """Checks if the job completed.

        :returns: True if job completed, otherwise False.
        :rtype: bool

        :raises vix.VixError: If failed to get job state.
        """

        result = ffi.new('Bool*')
        error_code = vix.VixJob_CheckCompletion(self._handle, result)
        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        return result[0]

    def get_error(self):
        """Gets an exception object.

        :returns: Exception object of job. The error may be VixError(VIX_OK).
        :rtype: .VixError
        """

        error_code = vix.VixJob_GetError(self._handle)
        return VixError(error_code)

    def _get_num_properties(self, property_id):
        count = vix.VixJob_GetNumProperties(
            self._handle,
            property_id,
        )
        return int(count)

    def _get_nth_properties(self, index, *args):

        c_args = list()
        for arg in args:
            alloc = None
            if arg in self.STR_RESULT_TYPES:
                alloc = ffi.new('char**')
            elif arg in (self.VIX_PROPERTY_JOB_RESULT_PROCESS_ID, self.VIX_PROPERTY_JOB_RESULT_FILE_SIZE, self.VIX_PROPERTY_JOB_RESULT_FILE_MOD_TIME):
                alloc = ffi.new('uint64*')
            else:
                alloc = ffi.new('int*')
            c_args.append(ffi.cast('VixPropertyType', arg))
            c_args.append(alloc)

        c_args.append(ffi.cast('VixPropertyType', self.VIX_PROPERTY_NONE))

        error_code = vix.VixJob_GetNthProperties(
            self._handle,
            index,
            *c_args
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        result = list()

        for i in range(len(args)):
            prop_id = int(c_args[i * 2])
            prop_val = c_args[(i * 2) + 1]

            value = None
            if prop_id in self.STR_RESULT_TYPES:
                value = _str(ffi.string(prop_val[0]), API_ENCODING)
                vix.Vix_FreeBuffer(prop_val[0])
            elif prop_id == self.VIX_PROPERTY_JOB_RESULT_PROCESS_BEING_DEBUGGED:
                value = bool(ffi.cast('Bool', prop_val[0]))
            elif prop_id in (self.VIX_PROPERTY_JOB_RESULT_PROCESS_START_TIME, self.VIX_PROPERTY_JOB_RESULT_FILE_MOD_TIME):
                value = datetime.datetime.fromtimestamp(int(ffi.cast('int', prop_val[0])))
            else:
                value = int(ffi.cast('int', prop_val[0]))

            result.append(value)

        return tuple(result)

    def get_properties(self, *args):
        """Get properties of a job result

        :param \*args: properties to fetch.

        :returns: A list of tuples of requests properties.
        :rtype: list

        :raises vix.VixError: On failure to fetch results.
        """

        num = self._get_num_properties(args[0])
        result = list()

        for i in range(num):
            result.append(self._get_nth_properties(i, *args))

        return result

    def __del__(self):
        self.release()



from __future__ import absolute_import

from .compat import _bytes, _str
from .VixError import VixError
from .VixHandle import VixHandle
from .VixSnapshot import VixSnapshot
from .VixJob import VixJob
from collections import namedtuple
from vix import _backend, API_ENCODING
vix = _backend._vix
ffi = _backend._ffi


def _blocking_job(f):
    def decorator(*args, **kwargs):
        job = f(*args, **kwargs)
        VixJob(job).wait()

    # allows sphinx to generate docs normally...
    decorator.__doc__ = f.__doc__

    return decorator

DirectoryListEntry = namedtuple('DirectoryListEntry', 'name size is_dir is_sym last_mod')
ProcessListEntry = namedtuple('ProcessListEntry', 'name pid owner cmd is_debug start_time')
SharedFolder = namedtuple('SharedFolder', 'name host_path write_access')

class VixVM(VixHandle):
    """Represents a guest VM."""
    
    _VIX_VMDELETE_DISK_FILES = 0x0002

    VIX_POWERSTATE_POWERING_OFF = 0x0001
    VIX_POWERSTATE_POWERED_OFF = 0x0002
    VIX_POWERSTATE_POWERING_ON = 0x0004
    VIX_POWERSTATE_POWERED_ON = 0x0008
    VIX_POWERSTATE_SUSPENDING = 0x0010
    VIX_POWERSTATE_SUSPENDED = 0x0020
    VIX_POWERSTATE_TOOLS_RUNNING = 0x0040
    VIX_POWERSTATE_RESETTING = 0x0080
    VIX_POWERSTATE_BLOCKED_ON_MSG = 0x0100
    VIX_POWERSTATE_PAUSED = 0x0200
    VIX_POWERSTATE_RESUMING = 0x0800

    VIX_TOOLSSTATE_UNKNOWN = 0x0001
    VIX_TOOLSSTATE_RUNNING = 0x0002
    VIX_TOOLSSTATE_NOT_INSTALLED = 0x0004

    VIX_VM_SUPPORT_SHARED_FOLDERS = 0x0001
    VIX_VM_SUPPORT_MULTIPLE_SNAPSHOTS = 0x0002
    VIX_VM_SUPPORT_TOOLS_INSTALL = 0x0004
    VIX_VM_SUPPORT_HARDWARE_UPGRADE = 0x0008

    _VIX_LOGIN_IN_GUEST_REQUIRE_INTERACTIVE_ENVIRONMENT = 0x08

    _VIX_RUNPROGRAM_RETURN_IMMEDIATELY = 0x0001
    _VIX_RUNPROGRAM_ACTIVATE_WINDOW = 0x0002

    VIX_VM_GUEST_VARIABLE = 1
    VIX_VM_CONFIG_RUNTIME_ONLY = 2
    VIX_GUEST_ENVIRONMENT_VARIABLE = 3

    _VIX_SNAPSHOT_REMOVE_CHILDREN = 0x0001

    _VIX_SNAPSHOT_INCLUDE_MEMORY = 0x0002

    _VIX_SHAREDFOLDER_WRITE_ACCESS = 0x04

    _VIX_CAPTURESCREENFORMAT_PNG = 0x01
    _VIX_CAPTURESCREENFORMAT_PNG_NOCOMPRESS = 0x02

    _VIX_CLONETYPE_FULL = 0
    _VIX_CLONETYPE_LINKED = 1

    _VIX_INSTALLTOOLS_MOUNT_TOOLS_INSTALLER = 0x00
    _VIX_INSTALLTOOLS_AUTO_UPGRADE = 0x01
    _VIX_INSTALLTOOLS_RETURN_IMMEDIATELY = 0x02

    VIX_VMPOWEROP_NORMAL = 0
    VIX_VMPOWEROP_FROM_GUEST = 0x0004
    VIX_VMPOWEROP_SUPPRESS_SNAPSHOT_POWERON = 0x0080
    VIX_VMPOWEROP_LAUNCH_GUI = 0x0200
    VIX_VMPOWEROP_START_VM_PAUSED = 0x1000

    def __init__(self, handle):
        super(VixVM, self).__init__(handle)

        assert self.get_type() == VixHandle.VIX_HANDLETYPE_VM, 'Expected VixVM handle.'

    @property
    def vmx_path(self):
        """Gets VM'x vmx path.

        :returns: Guest's VMX path.
        :rtype: str
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_VM_VMX_PATHNAME)

    @property
    def machine_info(self):
        """Get the VM's hardware specs

        :returns: a tuple: (num_cpus, memory)
        :rtype: tuple
        """

        return self.get_properties(
            VixHandle.VIX_PROPERTY_VM_NUM_VCPUS, 
            VixHandle.VIX_PROPERTY_VM_MEMORY_SIZE
        )

    @property
    def is_running(self):
        """Checks if VM is running

        :returns: True if running otherwise False.
        :rtype: bool
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_VM_IS_RUNNING)

    @property
    def guest_os(self):
        """The guest's Operating System

        :rtype: str
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_VM_GUESTOS)

    @property
    def name(self):
        """Name of the VM.

        :rtype: str
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_VM_NAME)

    @property
    def is_readonly(self):
        """Checks if the VM is readonly.

        :rtype: bool
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_VM_READ_ONLY)

    @property
    def power_state(self):
        """Gets the VMs power state.

        :returns: Any of VixVM.VIX_VM_POWERSTATE_*
        :rtype: int
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_VM_POWER_STATE)

    @property
    def tools_state(self):
        """Get the VMware tools state.

        :returns: Any of VixVM.VIX_VM_TOOLSSTATE_*
        :rtype: int
        """

        return self.get_properties(VixHandle.VIX_PROPERTY_VM_TOOLS_STATE)

    @property
    def supported_features(self):
        """Get features supported by the VM.

        :returns: Any of VixVM.VIX_VM_SUPPORT_*.
        :rtype: int
        """
        return self.get_properties(VixHandle.VIX_PROPERTY_VM_SUPPORTED_FEATURES)

    def __repr__(self):
        return "<VixVM @ {0}>".format(self.vmx_path)

    # Power
    @_blocking_job
    def pause(self):
        """Pauses the Virtual machine.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_Pause(
            self._handle,
            ffi.cast('int', 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def power_off(self, from_guest=False):
        """Powers off a VM.

        :param bool from_guest: True to initiate from guest, otherwise False.

        :raises vix.VixError: On failure to power off VM.
        """

        return vix.VixVM_PowerOff(
            self._handle,
            ffi.cast('VixVMPowerOpOptions', self.VIX_VMPOWEROP_FROM_GUEST if from_guest else self.VIX_VMPOWEROP_NORMAL),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def power_on(self, launch_gui=False):
        """Powers on a VM.

        :param bool launch_gui: True to launch GUI, otherwise False.

        :raises vix.VixError: On failure to power on VM.
        """

        return vix.VixVM_PowerOn(
            self._handle,
            ffi.cast('VixVMPowerOpOptions', self.VIX_VMPOWEROP_LAUNCH_GUI if launch_gui else self.VIX_VMPOWEROP_NORMAL),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def reset(self, from_guest=False):
        """Resets a virtual machine.

        :param bool from_guest: True to initiate from guest, otherwise False.

        :raises vix.VixError: On failure to reset VM.
        """

        return vix.VixVM_Reset(
            self._handle,
            ffi.cast('VixVMPowerOpOptions', self.VIX_VMPOWEROP_FROM_GUEST if from_guest else self.VIX_VMPOWEROP_NORMAL),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def suspend(self):
        """Suspends a virtual machine.

        :raises vix.VixError: On failure to suspend VM.
        """

        return vix.VixVM_Suspend(
            self._handle,
            ffi.cast('VixVMPowerOpOptions', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def unpause(self):
        """Resumes execution of a paused virtual machine.

        :raises vix.VixError: On failure to unpause VM.

        .. note:: This method is not supported by all Vmware products.
        """

        return vix.VixVM_Unpause(
            self._handle,
            ffi.cast('int', 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    # Snapshots
    def clone(self, dest_vmx, snapshot=None, linked=False):
        """Clones the VM to a specified location.

        :param str dest_vms: The clone will be stored here.
        :param .VixSnapshot snapshot: Optional snapshot as the state of the clone.
        :param bool linked: True for a linked clone, otherwise False for a full clone.

        :returns: Instance of the cloned VM.
        :rtype: .VixVM

        :raises vix.VixError: On failure to clone.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_Clone(
            self._handle,
            ffi.cast('VixHandle', snapshot._handle if snapshot else 0),
            ffi.cast('VixCloneType', self._VIX_CLONETYPE_LINKED if linked else self._VIX_CLONETYPE_FULL),
            ffi.new('char[]', _bytes(dest_vmx, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        return VixVM(job.wait(VixJob.VIX_PROPERTY_JOB_RESULT_HANDLE))

    def create_snapshot(self, name=None, description=None, include_memory=True):
        """Create a VM snapshot.

        :param str name: Name of snapshot.
        :param str description: Snapshot description.
        :param bool include_memory: True to include RAM, otherwise False.

        :returns: Instance of the created snapshot
        :rtype: .VixSnapshot

        :raises vix.VixError: On failure to create snapshot.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_CreateSnapshot(
            self._handle,
            ffi.new('char[]', _bytes(name, API_ENCODING)) if name else ffi.cast('char*', 0),
            ffi.new('char[]', _bytes(description, API_ENCODING)) if description else  ffi.cast('char*', 0),
            ffi.cast('int', self._VIX_SNAPSHOT_INCLUDE_MEMORY if include_memory else 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        return VixSnapshot(job.wait(VixJob.VIX_PROPERTY_JOB_RESULT_HANDLE))

    def snapshot_get_current(self):
        """Gets the VMs current active snapshot.

        :returns: The currently active snapshot
        :rtype: .VixSnapshot

        :raises vix.VixError: On failure to get the current snapshot.

        .. note:: This method is not supported by all VMware products.
        """

        snapshot_handle = ffi.new('VixHandle*')
        error_code = vix.VixVM_GetCurrentSnapshot(
            self._handle,
            snapshot_handle,
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        return VixSnapshot(snapshot_handle[0])

    def snapshot_get_named(self, name):
        """Gets a snapshot matching the given name.

        :param str name: Name of the snapshot to get.
        
        :returns: Instance of requests snapshot.
        :rtype: .VixSnapshot

        :raises vix.VixError: If failed to retreive desired snapshot.

        .. note:: This method is not supported by all VMware products.
        """

        snapshot_handle = ffi.new('VixHandle*')
        error_code = vix.VixVM_GetNamedSnapshot(
            self._handle,
            ffi.new('char[]', _bytes(name, API_ENCODING)),
            snapshot_handle,
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        return VixSnapshot(snapshot_handle[0])

    def snapshots_get_root_count(self):
        """Gets the count of root snapshots the VM owns.

        :returns: Count of VM's root snapshots.
        :rtype: int

        :raises vix.VixError: If failed to retrive root snapshot count.

        .. note:: This method is not supported by all VMware products.
        """

        result = ffi.new('int*')
        error_code = vix.VixVM_GetNumRootSnapshots(
            self._handle,
            result,
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        return result[0]

    def snapshot_get_root(self, index=0):
        """Gets the specified VM Snapshot.

        :param int index: zero based snapshot index.

        :returns: Root snapshot at specified index.
        :rtype: .VixSnapshot

        :raises vix.VixError: If failed to get specified too snapshot.

        .. note:: This methoid is not supported in all VMware products.
        """

        snapshot_handle = ffi.new('VixHandle*')
        error_code = vix.VixVM_GetRootSnapshot(
            self._handle,
            ffi.cast('int', index),
            snapshot_handle,
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        return VixSnapshot(snapshot_handle[0])

    @_blocking_job
    def snapshot_revert(self, snapshot, options=0):
        """Revet VM state to specified snapshot.

        :param .VixSnapshot snapshot: The snapshot to revert to.
        :param int options: Any of VIX_VMPOWEROP_*, VIX_VMPOWEROP_SUPPRESS_SNAPSHOT_POWERON is mutually exclusive.

        :raises vix.VixError: If failed to revert VM to snapshot.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_RevertToSnapshot(
            self._handle,
            snapshot._handle,
            ffi.cast('int', options),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def snapshot_remove(self, snapshot, remove_children=False):
        """Removed specified snapshot from VM.

        :param .VixSnapshot snapshot: The snapshot to remove.
        :param bool remove_children: True to remove child snapshots too, otherwise False.

        :raises vix.VixError: If failed to remove specified snapshot.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_RemoveSnapshot(
            self._handle,
            snapshot._handle,
            ffi.cast('int', self._VIX_SNAPSHOT_REMOVE_CHILDREN if remove_children else 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    # Guest & Host file mgmt.
    @_blocking_job
    def copy_guest_to_host(self, guest_path, host_path):
        """Copies a file or directory from the VM to host.

        :param str guest_path: Path to copy from on guest.
        :param str host_path: Path to copy to on host.

        :raises vix.VixError: If copy failed.
        """

        return vix.VixVM_CopyFileFromGuestToHost(
            self._handle,
            ffi.new('char[]', _bytes(guest_path, API_ENCODING)),
            ffi.new('char[]', _bytes(host_path, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def copy_host_to_guest(self, host_path, guest_path):
        """Copies a file or directory from host to VM.

        :param str host_path: Path to copy from on host.
        :param str guest_path: Path to copy to on VM.

        :raises vix.VixError: If failed to copy.
        """

        return vix.VixVM_CopyFileFromHostToGuest(
            self._handle,
            ffi.new('char[]', _bytes(host_path, API_ENCODING)),
            ffi.new('char[]', _bytes(guest_path, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def create_directory(self, path):
        """Creates a directory in the guest VM.

        :param str path: Path to create in guest.

        :raises vix.VixError: On failure to create directory.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_CreateDirectoryInGuest(
            self._handle,
            ffi.new('char[]', _bytes(path, API_ENCODING)),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    def create_temp(self):
        """Creates a temporary file in guest.

        :returns: Temporary file name.
        :rtype: str

        :raises vix.VixError: On failure to create temporary file.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_CreateTempFileInGuest(
            self._handle,
            ffi.cast('int', 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        return job.wait(VixJob.VIX_PROPERTY_JOB_RESULT_ITEM_NAME)

    @_blocking_job
    def file_rename(self, old_name, new_name):
        """Renames a file or directory in guest.

        :param str old_name: Name of file to rename.
        :param str new_name: The new name to give the file.

        :raises vix.VixError: On failure to rename.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_RenameFileInGuest(
            self._handle,
            ffi.new('char[]', _bytes(old_name, API_ENCODING)),
            ffi.new('char[]', _bytes(new_name, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def dir_delete(self, path):
        """Deletes a directory in guest VM.

        :param str path: Path of directory to delete.

        :raises vix.VixError: If failed to delete directory.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_DeleteDirectoryInGuest(
            self._handle,
            ffi.new('char[]', _bytes(path, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def file_delete(self, path):
        """Deletes a file in guest VM.

        :param str path: Path of file to delete.

        :raises vix.VixError: If failed to delete directory.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_DeleteFileInGuest(
            self._handle,
            ffi.new('char[]', _bytes(path, API_ENCODING)),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    def dir_exists(self, path):
        """Checks if a directory exists in guest VM.

        :param str path: Path to check if exists.

        :returns: True if directory exists, othwerwise False.
        :rtype: bool

        :raises vix.VixError: If failed to check.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_DirectoryExistsInGuest(
            self._handle,
            ffi.new('char[]', _bytes(path, API_ENCODING)),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        return bool(job.wait(VixJob.VIX_PROPERTY_JOB_RESULT_GUEST_OBJECT_EXISTS))

    def file_exists(self, path):
        """Checks if a file exists in guest VM.

        :param str path: File to check.

        :returns: True if file exists, otherwise False.
        :rtype: bool

        :raises vix.VixError: If failed to check file existance.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_FileExistsInGuest(
            self._handle,
            ffi.new('char[]', _bytes(path, API_ENCODING)),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        return bool(job.wait(VixJob.VIX_PROPERTY_JOB_RESULT_GUEST_OBJECT_EXISTS))

    def get_file_info(self, path):
        """Gets information about specified file in guest.

        :param str path: File path to get information about.

        :returns: DirectoryListEntry instance.
        :rtype: .DirectoryListEntry

        :raises vix.VixError: On failure to get file info.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_GetFileInfoInGuest(
            self._handle,
            ffi.new('char[]', _bytes(path, API_ENCODING)),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))
        job.wait()
        res = job.get_properties(
            VixJob.VIX_PROPERTY_JOB_RESULT_FILE_SIZE,
            VixJob.VIX_PROPERTY_JOB_RESULT_FILE_FLAGS,
            VixJob.VIX_PROPERTY_JOB_RESULT_FILE_MOD_TIME,
        )
        assert len(res) == 1, 'Expected single result'
        res = res[0]
        return DirectoryListEntry(
            name=None,
            size=res[0],
            is_dir=bool(res[1] & VixJob.VIX_FILE_ATTRIBUTES_DIRECTORY),
            is_sym=bool(res[1] & VixJob.VIX_FILE_ATTRIBUTES_SYMLINK),
            last_mod=res[2],
        )

    def dir_list(self, path):
        """Gets directory listing of specified path in guest VM.

        :param str path: Path to get directory list of.

        :returns: List of tuples, each containing: File Name, File Size, is dir, is symlink, mod time.

        :raises vix.VixError: On failure to get file info.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_ListDirectoryInGuest(
            self._handle,
            ffi.new('char[]', _bytes(path, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))
        job.wait()
        
        job_result = job.get_properties(
            VixJob.VIX_PROPERTY_JOB_RESULT_ITEM_NAME,
            VixJob.VIX_PROPERTY_JOB_RESULT_FILE_SIZE,
            VixJob.VIX_PROPERTY_JOB_RESULT_FILE_FLAGS,
            VixJob.VIX_PROPERTY_JOB_RESULT_FILE_MOD_TIME,
        )

        result = list()

        for res in job_result:
            result.append(DirectoryListEntry(
                name=res[0],
                size=res[1],
                is_dir=bool(res[2] & VixJob.VIX_FILE_ATTRIBUTES_DIRECTORY),
                is_sym=bool(res[2] & VixJob.VIX_FILE_ATTRIBUTES_SYMLINK),
                last_mod=res[3],
            ))

        return result

    # Guest execution
    @_blocking_job
    def proc_kill(self, pid):
        """Kills a process in guest VM.

        :param int pid: PID of process in guest to kill.

        :raises vix.VixError: If failed to kill process.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_KillProcessInGuest(
            self._handle,
            ffi.cast('uint64', pid),
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    def proc_list(self):
        """Gets the guest's process list.

        :returns: A list of tuples, each tuple contains: Process Name, PID, owner, cmd line, is debugged, start time.
        :rtype: list

        :raises vix.VixError: On failure to get file info.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_ListProcessesInGuest(
            self._handle,
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        job.wait()

        return [ProcessListEntry(
            name=entry[0],
            pid=entry[1],
            owner=entry[2],
            cmd=entry[3],
            is_debug=entry[4],
            start_time=entry[5],
        ) for entry in job.get_properties(
            VixJob.VIX_PROPERTY_JOB_RESULT_ITEM_NAME,
            VixJob.VIX_PROPERTY_JOB_RESULT_PROCESS_ID,
            VixJob.VIX_PROPERTY_JOB_RESULT_PROCESS_OWNER,
            VixJob.VIX_PROPERTY_JOB_RESULT_PROCESS_COMMAND,
            VixJob.VIX_PROPERTY_JOB_RESULT_PROCESS_BEING_DEBUGGED,
            VixJob.VIX_PROPERTY_JOB_RESULT_PROCESS_START_TIME,
        )]

    @_blocking_job
    def login(self, username, password, require_interactive=False):
        """Login to the guest to allow further executions.

        :param str username: Guest Login username.
        :param str password: Guest login password.
        :param bool require_interactive: If login requires an interactive session.

        :raises vix.VixError: On failure to authenticate.
        """

        return vix.VixVM_LoginInGuest(
            self._handle,
            ffi.new('char[]', _bytes(username, API_ENCODING)) if username else ffi.cast('char*', 0),
            ffi.new('char[]', _bytes(password, API_ENCODING)) if password else ffi.cast('char*', 0),
            ffi.cast('int', self._VIX_LOGIN_IN_GUEST_REQUIRE_INTERACTIVE_ENVIRONMENT if require_interactive else 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def logout(self):
        """Logout from guest. Closes any previous login context.

        :raises vix.VixError: If failed to logout.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_LogoutFromGuest(
            self._handle,
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def proc_run(self, program_name, command_line=None, should_block=True):
        """Executes a process in guest VM.

        :param str program_name: Name of program to execute in guest.
        :param str command_line: Command line to execute program with.
        :param bool should_block: If set to True, function will block until process exits in guest.

        :raises vix.VixError: On failure to execute process.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_RunProgramInGuest(
            self._handle,
            ffi.new('char[]', _bytes(program_name, API_ENCODING)),
            ffi.new('char[]', _bytes(command_line, API_ENCODING)) if command_line else ffi.cast('char*', 0),
            ffi.cast('VixRunProgramOptions', 0 if should_block else self._VIX_RUNPROGRAM_RETURN_IMMEDIATELY),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def run_script(self, script_text, interpreter_path=None, should_block=True):
        """Executes a script in guest VM.

        :param str script_text: The script to execute.
        :param str interpreter_path: Path of the interpreter for the script.
        :param bool should_block: If set to False, function will return immediately. True will block untill script returns.

        :raises vix.VixError: On failure to execute script.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_RunScriptInGuest(
            self._handle,
            ffi.new('char[]', _bytes(interpreter_path, API_ENCODING)) if interpreter_path else ffi.cast('char*', 0),
            ffi.new('char[]', _bytes(script_text, API_ENCODING)),
            ffi.cast('VixRunProgramOptions', 0 if should_block else self._VIX_RUNPROGRAM_RETURN_IMMEDIATELY),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    # Share mgmt.
    @_blocking_job
    def add_shared_folder(self, share_name, host_path, write_access=True):
        """Shares a folder with the guest VM.

        :param str share_name: Name of share in guest VM.
        :param str host_path: Path to share in host.
        :param bool write_access: True to allow guest to write to share.

        :raises vix.VixError: On failure to add share.

        .. note:: This method is not supported by all VMware products.
        """

        # TODO: return the path of shared folder in guest.
        return vix.VixVM_AddSharedFolder(
            self._handle,
            ffi.new('char[]', _bytes(share_name, API_ENCODING)),
            ffi.new('char[]', _bytes(host_path, API_ENCODING)),
            ffi.cast('VixMsgSharedFolderOptions', self._VIX_SHAREDFOLDER_WRITE_ACCESS if write_access else 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def share_enable(self, value=True):
        """Enables/Disables shares between Host and guest VM.

        :param bool value: True to enable, False to disable.

        :raises vix.VixError: If failed to enable/disable shares.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_EnableSharedFolders(
            self._handle,
            ffi.cast('Bool', int(value)),
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    def get_shared_folder_count(self):
        """Gets the count of shared folder of VM with host.

        .. note:: This method is not supported by all VMware products.
        """

        return VixJob(vix.VixVM_GetNumSharedFolders(
            self._handle,
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )).wait(VixJob.VIX_PROPERTY_JOB_RESULT_SHARED_FOLDER_COUNT)

    def get_shared_folder_state(self, index):
        """Gets the state of a shared folder.

        :param int index: Index of share.

        :returns: tuple with share state information.
        :rtype: SharedFolder

        :raises vix.VixError: If failed to get state.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_GetSharedFolderState(
            self._handle,
            ffi.cast('int', index),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))
        job.wait()
        res = job.get_properties(
            VixJob.VIX_PROPERTY_JOB_RESULT_ITEM_NAME,
            VixJob.VIX_PROPERTY_JOB_RESULT_SHARED_FOLDER_HOST,
            VixJob.VIX_PROPERTY_JOB_RESULT_SHARED_FOLDER_FLAGS,
        )
        assert len(res) == 1, 'Expected single result'
        res = res[0]
        return SharedFolder(
            name=res[0],
            host_path=res[1],
            write_access=bool(res[2] & VixVM._VIX_SHAREDFOLDER_WRITE_ACCESS),
        )

    @_blocking_job
    def share_remove(self, share_name):
        """Removes a share between host and guest VM.

        :param str share_name: Name of share to remove.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_RemoveSharedFolder(
            self._handle,
            ffi.new('char[]', _bytes(share_name, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def share_set_state(self, share_name, host_path, allow_write=True):
        """Sets the state for an existing share.

        :param str share_name: Name of share to modify.
        :param str host_path: Path on host to set to share.
        :param bool allow_write: Sets if the guest will be able to write to share.

        :raises vix.VixError: If failed to set share state.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_SetSharedFolderState(
            self._handle,
            ffi.new('char[]', _bytes(share_name, API_ENCODING)),
            ffi.new('char[]', _bytes(host_path, API_ENCODING)),
            ffi.cast('VixMsgSharedFolderOptions', VixVM._VIX_SHAREDFOLDER_WRITE_ACCESS if allow_write else 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    # VM environment.
    def var_read(self, name, variable_type=VIX_VM_GUEST_VARIABLE):
        """Reads an environment string.

        :param str name: Name of variable to read.
        :param int variable_type: Must be one of VIX_VM_GUEST_VARIABLE, VIX_VM_CONFIG_RUNTIME_ONLY or VIX_GUEST_ENVIRONMENT_VARIABLE.

        :returns: The value of the variable.
        :rtype: str

        :raises vix.VixError: On failure to get specified variable.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixVM_ReadVariable(
            self._handle,
            ffi.cast('int', variable_type),
            ffi.new('char[]', _bytes(name, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        return job.wait(VixJob.VIX_PROPERTY_JOB_RESULT_VM_VARIABLE_STRING)

    @_blocking_job
    def var_write(self, name, value, variable_type=VIX_VM_GUEST_VARIABLE):
        """Writes a string to the VM's environment.

        :param str name: Name of env string to set.
        :param str value: Value of env string to set.
        :param int variable_type: Must be one of VIX_VM_GUEST_VARIABLE, VIX_VM_CONFIG_RUNTIME_ONLY or VIX_GUEST_ENVIRONMENT_VARIABLE.

        :raises vix.VixError: On failure to set environment.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_WriteVariable(
            self._handle,
            ffi.cast('int', variable_type),
            ffi.new('char[]', _bytes(name, API_ENCODING)),
            ffi.new('char[]', _bytes(value, API_ENCODING)),
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    # Misc. methods
    @_blocking_job
    def upgrade_virtual_hardware(self):
        """Upgrades virtual hardware of a virtual machine.

        :raises vix.VixError: If failed to upgrade virtual hardware.

        .. note:: This method is not supported by all VMware products.
        """

        return vix.VixVM_UpgradeVirtualHardware(
            self._handle,
            ffi.cast('int', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def vm_delete(self, delete_files=False):
        """Deletes VM from host.

        :param bool delete_files: True to delete associated files from disk, otherwise False.

        :raises vix.VixError: If failed to delete VM.
        """

        return vix.VixVM_Delete(
            self._handle,
            ffi.cast('VixVMDeleteOptions', self._VIX_VMDELETE_DISK_FILES if delete_files else 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    def capture_screen_image(self, filename=None):
        """Captures a PNG screenshot from VM.

        :param str filename: A filename to save image to. If this file exists it will be overwritten. None will return the image a string/BLOB.

        :returns: A PNG image in binary form if filename is None, otherwise None
        :rtype: bytes

        :raises vix.VixError: On failure to capture screenshot.

        .. note:: This method is not supported by all VMWare products.
        """

        job = VixJob(vix.VixVM_CaptureScreenImage(
            self._handle,
            ffi.cast('int', self._VIX_CAPTURESCREENFORMAT_PNG),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        bytes_ptr = ffi.new('int*')
        data_ptr = ffi.new('char**')

        error_code = vix.VixJob_Wait(
            job._handle,
            ffi.cast('int', VixJob.VIX_PROPERTY_JOB_RESULT_SCREEN_IMAGE_DATA),
            bytes_ptr,
            data_ptr,
            ffi.cast('int', VixJob.VIX_PROPERTY_NONE)
        )

        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

        img_len = int(bytes_ptr[0])
        img_data = _bytes(ffi.buffer(data_ptr[0], img_len))

        vix.Vix_FreeBuffer(data_ptr[0])

        if filename:
            with open(filename, "wb") as fd:
                fd.write(img_data)
        else:
            return img_data

    # VMware tools
    @_blocking_job
    def wait_for_tools(self, timeout=0):
        """Waits for VMware tools to start in guest.

        :param int timeout: Timeout in seconds. Zero or negative will block forever, Raises an exception if timeout expired.

        :raises vix.VixError: If timeout passed, Or of VIX fails.
        """

        return vix.VixVM_WaitForToolsInGuest(
            self._handle,
            ffi.cast('int', timeout),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    @_blocking_job
    def install_tools(self, auto_upgrade=False, blocking=True):
        """Starts the VMware tools install operation on guest.

        :param bool auto_upgrade: True for auto upgrade, False will mount installer.
        :param bool blocking: False will return immediatly, True will wait till operation ends.

        :raises vix.VixError: On failure to start install.
        """

        options = 0
        if blocking:
            options |= self._VIX_INSTALLTOOLS_RETURN_IMMEDIATELY
        if auto_upgrade:
            options |= self._VIX_INSTALLTOOLS_AUTO_UPGRADE

        return vix.VixVM_InstallTools(
            self._handle,
            ffi.cast('int', options),
            ffi.cast('char*', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

    def __del__(self):
        self.release()



from __future__ import absolute_import

import collections
import platform
import os
import os.path
from xml.etree import ElementTree
from .compat import _bytes, _str
from .VixJob import VixJob
from .VixVM import VixVM
from .VixHandle import VixHandle
from .VixError import VixError
from vix import _backend, API_ENCODING
vix = _backend._vix
ffi = _backend._ffi

_find_results = dict()

@ffi.callback('void(*)(VixHandle, VixEventType, VixHandle, void*)')
def _find_items_callback(job_handle, event_type, event_info, client_data):
    """Callback that collects found VMs.

    .. note:: Internal use.
    """

    VIX_EVENTTYPE_JOB_COMPLETED = 2
    VIX_EVENTTYPE_JOB_PROGRESS = 3
    VIX_EVENTTYPE_FIND_ITEM = 8

    if event_type != VIX_EVENTTYPE_FIND_ITEM:
        return

    idx = int(ffi.cast('int', client_data))
    str_ptr = ffi.new('char**')

    error_code = vix.Vix_GetProperties(
        ffi.cast('VixHandle', event_info),
        ffi.cast('VixPropertyID', VixJob.VIX_PROPERTY_FOUND_ITEM_LOCATION),
        ffi.cast('char**', str_ptr),
        ffi.cast('VixPropertyID', VixJob.VIX_PROPERTY_NONE),
    )

    if error_code != VixError.VIX_OK:
        return

    vmx = _str(ffi.string(str_ptr[0]), API_ENCODING)

    _find_results[idx].append(vmx)
    vix.Vix_FreeBuffer(str_ptr[0])

VixHostInfo = collections.namedtuple('VixHostInfo', 'host_type api_version software_version')


class VixHost(object):
    """Represents a VMware virtualization host."""

    _VIX_API_VERSION = -1

    _VIX_HOSTOPTION_VERIFY_SSL_CERT = 0x4000

    VIX_SERVICEPROVIDER_DEFAULT = 1
    VIX_SERVICEPROVIDER_VMWARE_SERVER = 2
    VIX_SERVICEPROVIDER_VMWARE_WORKSTATION = 3
    VIX_SERVICEPROVIDER_VMWARE_PLAYER = 4
    VIX_SERVICEPROVIDER_VMWARE_VI_SERVER = 10
    VIX_SERVICEPROVIDER_VMWARE_WORKSTATION_SHARED = 11

    VIX_FIND_RUNNING_VMS = 1
    VIX_FIND_REGISTERED_VMS = 4

    @property
    def host_info(self):
        """property returns the Host's version.

        :returns: Host's type, api and software version.
        :rtype: .VixHostInfo

        :raises vix.VixError: If failed to get version information.
        """
        res = VixHandle(self._handle).get_properties(
            VixHandle.VIX_PROPERTY_HOST_HOSTTYPE,
            VixHandle.VIX_PROPERTY_HOST_API_VERSION,
            VixHandle.VIX_PROPERTY_HOST_SOFTWARE_VERSION,
        )
        return VixHostInfo(host_type=res[0], api_version=res[1], software_version=res[2])

    def __init__(self, service_provider=VIX_SERVICEPROVIDER_DEFAULT, host=None, credentials=None):
        """Connects to a VMware host.

        :param int service_provider: Specifies the service to connect to, may be any of VIX_SERVICEPROVIDER_* constants.
        :param tuple host: (hostname, port).
        :param tuple credentials: (username, password).

        :raises AssertionError: If passed arguments are not the right types/sizes.
        :raises vix.VixError: On connection failure.
        """
        self._handle = None

        assert self._handle == None, 'Instance is already connected.'

        if service_provider in (self.VIX_SERVICEPROVIDER_VMWARE_VI_SERVER, self.VIX_SERVICEPROVIDER_VMWARE_WORKSTATION_SHARED):
            if isinstance(host, collections.Sequence) and len(host) == 2 and host[1] != 0:
                host = ('%s:%s' % host, 0)

            elif host is None and service_provider == self.VIX_SERVICEPROVIDER_VMWARE_WORKSTATION_SHARED:
                if platform.system() == 'Windows':
                    vmhostd_proxy_xml = os.path.join(os.getenv('ProgramData'), 'VMware', 'hostd', 'proxy.xml')
                elif platform.system() == 'Linux':
                    vmhostd_proxy_xml = '/etc/vmware/hostd/proxy.xml'
                else:
                    raise NotImplemented('Unrecognized OS or architecure ({0}, {1}, {2})'.format(platform.architecure(), platform.machine(), platform.system()))

                host = ('localhost:%d' % (int(ElementTree.parse(vmhostd_proxy_xml).getroot().find('httpsPort').text), ), 0)

            assert credentials is not None and isinstance(credentials, collections.Sequence) and len(credentials) == 2 \
                   and credentials[0] is not None and credentials[1] is not None, \
                   'Null credentials blocks vix api.'

        if not host:
            host = (None, 0, )
        if not credentials:
            credentials = (None, None, )

        assert isinstance(host, collections.Sequence) and len(host) == 2, 'Host must be a tuple of size 2'
        assert isinstance(credentials, collections.Sequence) and len(credentials) == 2, 'Credentials must be a tuple of size 2'

        job = vix.VixHost_Connect(
            self._VIX_API_VERSION,
            service_provider,
            ffi.from_buffer(_bytes(host[0], API_ENCODING)) if host[0] else ffi.cast('const char*', 0),
            host[1],
            ffi.from_buffer(_bytes(credentials[0], API_ENCODING)) if credentials[0] is not None else ffi.cast('const char*', 0),
            ffi.from_buffer(_bytes(credentials[1], API_ENCODING)) if credentials[1] is not None else ffi.cast('const char*', 0),
            0,
            0,
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        )

        if VixHandle(job).get_type() != VixHandle.VIX_HANDLETYPE_JOB:
            raise VixError(vix.VixJob_GetError(job))

        job = VixJob(job)

        self._handle = job.wait(VixJob.VIX_PROPERTY_JOB_RESULT_HANDLE)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.disconnect()

    def disconnect(self):
        """Disconnects from the host."""

        if self._handle:
            vix.VixHost_Disconnect(self._handle)
            VixHandle(self._handle).release()
            self._handle = None

    def register_vm(self, vmx_path):
        """Registers a VM to host.

        :param str vmx_path: Path of VM configuration to register.

        :raises vix.VixError: If failed to register VM.
        
        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixHost_RegisterVM(
            self._handle,
            ffi.cast('const char*', _bytes(vmx_path, API_ENCODING)),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))
        error_code = job.wait()
        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

    def unregister_vm(self, vmx_path):
        """Unregisters a VM from host.

        :param str vmx_path: Path of VM configuration to unregister.
        
        :raises vix.VixError: If failed to unregister VM.

        .. note:: This method is not supported by all VMware products.
        """

        job = VixJob(vix.VixHost_UnregisterVM(
            self._handle,
            ffi.cast('const char*', _bytes(vmx_path, API_ENCODING)),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))
        error_code = job.wait()
        if error_code != VixError.VIX_OK:
            raise VixError(error_code)

    def open_vm(self, vmx_path):
        """Opens a VM in host.

        :param str vmx_path: Path to requested VM's configuration.

        :rtype: vix.VixVM

        :raises vix.VixError: If failed to open requested VM.
        """

        assert self._handle is not None, 'Must be connected.'

        job = VixJob(vix.VixHost_OpenVM(
            self._handle,
            ffi.new('char[]', _bytes(vmx_path, API_ENCODING)),
            ffi.cast('VixVMOpenOptions', 0),
            ffi.cast('VixHandle', 0),
            ffi.cast('VixEventProc*', 0),
            ffi.cast('void*', 0),
        ))

        return VixVM(job.wait(VixJob.VIX_PROPERTY_JOB_RESULT_HANDLE))

    def find_items(self, search_type=VIX_FIND_RUNNING_VMS, names_only=False):
        """Finds VMs on host with requested citeria.

        :param int search_type: Any of VIX_FIND_*.
        :param bool names_only: True will return a list of vmx paths, False will return VixVM instances.

        :returns: List of found VMs.
        :rtype: list

        :raises vix.VixError: On failure to find items.
        """

        key = len(_find_results)
        _find_results[key] = list()

        job = VixJob(vix.VixHost_FindItems(
            self._handle,
            ffi.cast('VixFindItemType', search_type),
            ffi.cast('VixHandle', 0),
            ffi.cast('int32', -1),
            ffi.cast('VixEventProc*', _find_items_callback),
            ffi.cast('void*', key),
        ))

        job.wait()

        items = _find_results.pop(key)

        if names_only:
            return items
        else:
            return [self.open_vm(vmx_path) for vmx_path in items]

    def __del__(self):
        self.disconnect()



import os
import cffi
import platform
import sys


class VixBackend(object):
    """This class creates the ffi and vix objects that will be used by other classes of this module.

    .. note:: Internal use.
    """
    def __init__(self):
        self._ffi = cffi.FFI()

        self._ffi.cdef('''
// Basic Types
typedef char          Bool;
typedef uint64_t    uint64;
typedef  int64_t     int64;
typedef uint32_t    uint32;
typedef  int32_t     int32;
typedef uint16_t    uint16;
typedef  int16_t     int16;
typedef  uint8_t    uint8;
typedef   int8_t     int8;

// VIX Types
typedef int VixHandle;
typedef int VixHandleType;
typedef uint64 VixError;
typedef int VixPropertyType;
typedef int VixPropertyID;
typedef int VixEventType;
typedef int VixHostOptions;
typedef int VixServiceProvider;
typedef int VixFindItemType;
typedef int VixVMOpenOptions;
typedef int VixVMPowerOpOptions;
typedef int VixVMDeleteOptions;
typedef int VixPowerState;
typedef int VixToolsState;
typedef int VixRunProgramOptions;
typedef int VixRemoveSnapshotOptions;
typedef int VixCreateSnapshotOptions;
typedef int VixMsgSharedFolderOptions;
typedef int VixCloneType;
typedef void VixEventProc(VixHandle handle, VixEventType eventType, VixHandle moreEventInfo, void *clientData);

// VIX functions
const char *Vix_GetErrorText(VixError err, const char *locale);

// Vix Handle functions
void Vix_ReleaseHandle(VixHandle handle);
void Vix_AddRefHandle(VixHandle handle);
VixHandleType Vix_GetHandleType(VixHandle handle);
VixError Vix_GetProperties(VixHandle handle, VixPropertyID firstPropertyID, ...);
VixError Vix_GetPropertyType(VixHandle handle, VixPropertyID propertyID, VixPropertyType *propertyType);
void Vix_FreeBuffer(void *p);

// Vix Host
VixHandle VixHost_Connect(int apiVersion, VixServiceProvider hostType, const char *hostName, int hostPort, const char *userName, const char *password, VixHostOptions options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData); 
void VixHost_Disconnect(VixHandle hostHandle);

// VM Registration
VixHandle VixHost_RegisterVM(VixHandle hostHandle, const char *vmxFilePath, VixEventProc *callbackProc, void *clientData);
VixHandle VixHost_UnregisterVM(VixHandle hostHandle, const char *vmxFilePath, VixEventProc *callbackProc, void *clientData);

// VM search
VixHandle VixHost_FindItems(VixHandle hostHandle, VixFindItemType searchType, VixHandle searchCriteria, int32 timeout, VixEventProc *callbackProc, void *clientData);

VixHandle VixHost_OpenVM(VixHandle hostHandle, const char *vmxFilePathName, VixVMOpenOptions options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);

// Property list
VixError VixPropertyList_AllocPropertyList(VixHandle hostHandle, VixHandle *resultHandle,  int firstPropertyID, ...);

// Vix VM
VixHandle VixVM_PowerOn(VixHandle vmHandle, VixVMPowerOpOptions powerOnOptions, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_PowerOff(VixHandle vmHandle, VixVMPowerOpOptions powerOffOptions, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_Reset(VixHandle vmHandle, VixVMPowerOpOptions resetOptions, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_Suspend(VixHandle vmHandle, VixVMPowerOpOptions suspendOptions, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_Pause(VixHandle vmHandle, int options, VixHandle propertyList, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_Unpause(VixHandle vmHandle, int options, VixHandle propertyList, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_Delete(VixHandle vmHandle, VixVMDeleteOptions deleteOptions, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_WaitForToolsInGuest(VixHandle vmHandle, int timeoutInSeconds, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_LoginInGuest(VixHandle vmHandle, const char *userName, const char *password, int options, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_LogoutFromGuest(VixHandle vmHandle, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_RunProgramInGuest(VixHandle vmHandle, const char *guestProgramName, const char *commandLineArgs, VixRunProgramOptions options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_ListProcessesInGuest(VixHandle vmHandle, int options, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_KillProcessInGuest(VixHandle vmHandle, uint64 pid, int options, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_RunScriptInGuest(VixHandle vmHandle, const char *interpreter, const char *scriptText, VixRunProgramOptions options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_CopyFileFromHostToGuest(VixHandle vmHandle, const char *hostPathName, const char *guestPathName, int options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_CopyFileFromGuestToHost(VixHandle vmHandle, const char *guestPathName, const char *hostPathName, int options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_DeleteFileInGuest(VixHandle vmHandle, const char *guestPathName, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_FileExistsInGuest(VixHandle vmHandle, const char *guestPathName, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_RenameFileInGuest(VixHandle vmHandle, const char *oldName, const char *newName, int options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_CreateTempFileInGuest(VixHandle vmHandle, int options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_GetFileInfoInGuest(VixHandle vmHandle, const char *pathName, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_ListDirectoryInGuest(VixHandle vmHandle, const char *pathName, int options, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_CreateDirectoryInGuest(VixHandle vmHandle, const char *pathName, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_DeleteDirectoryInGuest(VixHandle vmHandle, const char *pathName, int options, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_DirectoryExistsInGuest(VixHandle vmHandle, const char *pathName, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_ReadVariable(VixHandle vmHandle, int variableType, const char *name, int options, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_WriteVariable(VixHandle vmHandle, int variableType, const char *valueName, const char *value, int options, VixEventProc *callbackProc, void *clientData);

VixError VixVM_GetNumRootSnapshots(VixHandle vmHandle, int *result);
VixError VixVM_GetRootSnapshot(VixHandle vmHandle, int index, VixHandle *snapshotHandle);
VixError VixVM_GetCurrentSnapshot(VixHandle vmHandle, VixHandle *snapshotHandle);
VixError VixVM_GetNamedSnapshot(VixHandle vmHandle, const char *name,VixHandle *snapshotHandle);
VixHandle VixVM_RemoveSnapshot(VixHandle vmHandle, VixHandle snapshotHandle, VixRemoveSnapshotOptions options, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_RevertToSnapshot(VixHandle vmHandle, VixHandle snapshotHandle, VixVMPowerOpOptions options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_CreateSnapshot(VixHandle vmHandle, const char *name, const char *description, VixCreateSnapshotOptions options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_EnableSharedFolders(VixHandle vmHandle, Bool enabled,       int options,        VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_GetNumSharedFolders(VixHandle vmHandle, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_GetSharedFolderState(VixHandle vmHandle, int index, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_SetSharedFolderState(VixHandle vmHandle, const char *shareName, const char *hostPathName, VixMsgSharedFolderOptions flags, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_AddSharedFolder(VixHandle vmHandle, const char *shareName, const char *hostPathName, VixMsgSharedFolderOptions flags, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_RemoveSharedFolder(VixHandle vmHandle, const char *shareName, int flags, VixEventProc *callbackProc, void *clientData);

VixHandle VixVM_CaptureScreenImage(VixHandle vmHandle, int captureType, VixHandle additionalProperties, VixEventProc *callbackProc, void *clientdata);

VixHandle VixVM_Clone(VixHandle vmHandle, VixHandle snapshotHandle, VixCloneType cloneType, const char *destConfigPathName, int options, VixHandle propertyListHandle, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_UpgradeVirtualHardware(VixHandle vmHandle, int options, VixEventProc *callbackProc, void *clientData);
VixHandle VixVM_InstallTools(VixHandle vmHandle, int options, const char *commandLineArgs, VixEventProc *callbackProc, void *clientData);

VixError VixJob_Wait(VixHandle jobHandle,  VixPropertyID firstPropertyID,  ...);
VixError VixJob_CheckCompletion(VixHandle jobHandle, Bool *complete);
VixError VixJob_GetError(VixHandle jobHandle);

int VixJob_GetNumProperties(VixHandle jobHandle, int resultPropertyID);
VixError VixJob_GetNthProperties(VixHandle jobHandle, int index, int propertyID, ...);

VixError VixSnapshot_GetNumChildren(VixHandle parentSnapshotHandle, int *numChildSnapshots);
VixError VixSnapshot_GetChild(VixHandle parentSnapshotHandle, int index, VixHandle *childSnapshotHandle);
VixError VixSnapshot_GetParent(VixHandle snapshotHandle, VixHandle *parentSnapshotHandle);
        ''')

        # Don't fail build for documentation.
        self._vix = self._ffi.dlopen(VixBackend._get_vix_path()) if os.getenv('READTHEDOCS') is None else None

    @staticmethod
    def _get_vix_path():
        """Finds the expected install path for the so/dll.

        :returns: Expected path.
        :rtype: str

        :raises NotImplemented: If the OS/architecture are not recognized.
        """

        os_name = platform.system()

        if os_name == 'Linux':
            return '/usr/lib/libvixAllProducts.so'

        elif os_name == 'Darwin':
            return '/Applications/VMware Fusion.app/Contents/Public/libvixAllProducts.dylib'

        elif os_name == 'Windows':
            arch = platform.architecture()[0].lower()
            machine = platform.machine().lower()

            # Py: 32; Machine: 32: %ProgramFiles%\VMware\VMware VIX\VixAllProductsDyn.dll
            # Py: 32; Machine: 64: %ProgramFiles(x86)%\VMware\VMware VIX\VixAllProductsDyn.dll
            # Py: 64; Machine: 32: N/A
            # Py: 64; Machine: 64: %ProgramFiles(x86)%\VMware\VMware VIX\Vix64AllProductsDyn.dll

            from six.moves import winreg

            try:
                if machine == 'amd64':
                    base_path = winreg.QueryValueEx(winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r'SOFTWARE\WOW6432Node\VMware, Inc.\VMware VIX'), 'InstallPath')[0]
                else:
                    base_path = winreg.QueryValueEx(winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r'SOFTWARE\VMware, Inc.\VMware VIX'), 'InstallPath')[0]
            except OSError:
                base_path = os.getenv('ProgramFiles(x86)') if machine == 'amd64' else os.getenv('ProgramFiles')
                base_path = os.path.join(base_path, 'VMware', 'VMware VIX')

            base_path = os.path.join(base_path, 'Vix64AllProductsDyn.dll' if arch == '64bit' else 'VixAllProductsDyn.dll')

            return base_path

        raise NotImplementedError('Unrecognized OS or architecture ({0}, {1}, {2})'.format(platform.architecture(), platform.machine(), platform.system()))



import os
import re
from setuptools import setup

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(BASE_DIR, 'README.rst'), 'rt') as fd:
    long_description = fd.read()

with open(os.path.join(BASE_DIR, 'vix', '__init__.py'), 'rt') as fd:
    version = re.search(r'__version__\W*=\W*(\'|")(.+?)(\'|")', fd.read()).group(2)

setup(
    name='vix',
    version=version,
    description='VMware VIX binding for Python (unofficial)',
    long_description=long_description,
    author='Naim A.',
    author_email='naim94a@gmail.com',
    url='https://github.com/naim94a/vix',
    packages=['vix'],
    package_dir={
        'vix': 'vix',
    },
    license='GPLv3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: System :: Systems Administration',
    ],
    install_requires=[
        'cffi>=1.8.2',
        'six',
    ],
    keywords='vmware python api vix',
    project_urls={
        'Documentation': 'https://naim94a.github.io/vix',
        'Source': 'https://github.com/naim94a/vix',
        'Bugs & Features': 'https://github.com/naim94a/vix/issues'
    }
)
